!*******************************************************************************

      module xml_parser

      use characters
      use strings

      private
      public parse_document
      integer, parameter, public :: HANDLER_ERROR                 =  0
      integer, parameter, public :: HANDLER_START_DOCUMENT        =  1
      integer, parameter, public :: HANDLER_START_PROLOG          =  2
      integer, parameter, public :: HANDLER_START_XML_DECLARATION =  3
      integer, parameter, public :: HANDLER_END_XML_DECLARATION   =  4
      integer, parameter, public :: HANDLER_DTD                   =  5
      integer, parameter, public :: HANDLER_END_PROLOG            =  6
      integer, parameter, public :: HANDLER_START_ROOT_ELEMENT    =  7
      integer, parameter, public :: HANDLER_START_ELEMENT         =  8
      integer, parameter, public :: HANDLER_START_ATTRIBUTES      =  9
      integer, parameter, public :: HANDLER_ATTRIBUTE             = 10
      integer, parameter, public :: HANDLER_END_ATTRIBUTES        = 11
      integer, parameter, public :: HANDLER_EMPTY_ELEMENT         = 12
      integer, parameter, public :: HANDLER_START_CONTENT         = 13
      integer, parameter, public :: HANDLER_CHARDATA              = 14
      integer, parameter, public :: HANDLER_PI                    = 15
      integer, parameter, public :: HANDLER_COMMENT               = 16
      integer, parameter, public :: HANDLER_CDATA                 = 17
      integer, parameter, public :: HANDLER_REFERENCE             = 18
      integer, parameter, public :: HANDLER_END_CONTENT           = 19
      integer, parameter, public :: HANDLER_END_ELEMENT           = 20
      integer, parameter, public :: HANDLER_END_ROOT_ELEMENT      = 21
      integer, parameter, public :: HANDLER_END_DOCUMENT          = 22

      contains

!*******************************************************************************
      subroutine parse_document(buffer,handler,c,status)

      implicit none

      interface
          subroutine handler(in,name,value,out)
          implicit none
          integer, intent(in)       :: in
          character(*), intent(in)  :: name,value
          integer, intent(out)      :: out
          end subroutine handler
      end interface

      character(*), intent(in)  :: buffer
      integer, intent(out)      :: status
      integer, intent(out)      :: c
      integer                   :: h

!     external handler


      call handler(HANDLER_START_DOCUMENT,'','',h)

      c = 1
      call handler(HANDLER_START_PROLOG,'','',h)
      call parse_prolog(buffer,c,handler,status)
      if (status /= 0) return
      call handler(HANDLER_END_PROLOG,'','',h)
      call handler(HANDLER_START_ROOT_ELEMENT,'','',h)
      call parse_element(buffer,c,handler,status)
      if (status /= 0) return
      call handler(HANDLER_END_ROOT_ELEMENT,'','',h)
      call parse_misc(buffer,c,handler,status)
      if (status /= 0) return

      call handler(HANDLER_END_DOCUMENT,'','',h)

      end subroutine parse_document
!*******************************************************************************
      subroutine parse_prolog(buffer,c,handler,status)

      implicit none
      character(*), intent(in) :: buffer
      integer, intent(inout) :: c
      integer, intent(out) :: status
      integer :: l

      external handler


      call parse_xml_declaration(buffer,c,handler,status)
      if (status /= 0) return

      l = len(buffer)
      do while (c <= l)
          if (is_space(cbuffer(buffer,c,c))) then
              call parse_misc(buffer,c,handler,status)
              if (status /= 0) return
          elseif (cbuffer(buffer,c,c+1) == '<?') then
              call parse_misc(buffer,c,handler,status)
              if (status /= 0) return
          elseif (cbuffer(buffer,c,c+3) == '<!--') then
              call parse_misc(buffer,c,handler,status)
              if (status /= 0) return
          elseif (cbuffer(buffer,c,c+2) == '<!D') then
              call parse_dtd(buffer,c,handler,status)
              if (status /= 0) return
          elseif (cbuffer(buffer,c,c) == '<') then
              exit
          else
              status = -1
              return
          endif
      enddo

      end subroutine parse_prolog
!*******************************************************************************
      subroutine parse_misc(buffer,c,handler,status)

      implicit none
      character(*), intent(in) :: buffer
      integer, intent(inout) :: c
      integer, intent(out) :: status
      integer :: l

      external handler


      l = len(buffer)
      do while (c <= l)
          if (is_space(cbuffer(buffer,c,c))) then
              call parse_space(buffer,c,status)
              if (status /= 0) return
          elseif (cbuffer(buffer,c,c+1) == '<?') then
              call parse_pi(buffer,c,handler,status)
              if (status /= 0) return
          elseif (cbuffer(buffer,c,c+3) == '<!--') then
              call parse_comment(buffer,c,handler,status)
              if (status /= 0) return
          else
              exit
          endif
      enddo

      end subroutine parse_misc
!*******************************************************************************
      recursive subroutine parse_element(buffer,c,handler,status)

      implicit none
      character(*), intent(in) :: buffer
      integer, intent(inout) :: c
      integer, intent(out) :: status
      type(string) :: start_name,end_name
      integer :: h,l
      logical :: empty

      external handler


!***  Start tag: name of element and attributes
      if (cbuffer(buffer,c,c) /= '<') then
          status = -14
          goto 9999
      endif
      call parse_element_starttag(buffer,c,handler,  &
               start_name,empty,status)
      if (status /= 0) goto 9999

!***  Empty element:
      if (empty) then
          goto 9999
      endif

!***  Content for (apparently) non-empty elements:
      call handler(HANDLER_START_CONTENT,'','',h)
      l = len(buffer)
      do while (c < l)
          if (cbuffer(buffer,c,c+1) == '</') then
              exit
          elseif (cbuffer(buffer,c,c+3) == '<!--') then
              call parse_comment(buffer,c,handler,status)
              if (status /= 0) goto 9999
          elseif (cbuffer(buffer,c,c+1) == '<?') then
              call parse_pi(buffer,c,handler,status)
              if (status /= 0) goto 9999
          elseif (cbuffer(buffer,c,c) == '&') then
              call parse_reference(buffer,c,handler,status)
              if (status /= 0) goto 9999
          elseif (cbuffer(buffer,c,c+8) == '<![CDATA[') then
              call parse_cdata(buffer,c,handler,status)
              if (status /= 0) goto 9999
          elseif (cbuffer(buffer,c,c) == '<') then
              call parse_element(buffer,c,handler,status)
              if (status /= 0) goto 9999
          else
              call parse_chardata(buffer,c,handler,status)
              if (status /= 0) goto 9999
          endif
      enddo
      call handler(HANDLER_END_CONTENT,'','',h)

!***  Endtag of element
      call parse_element_endtag(buffer,c,handler,end_name,status)
      if (status /= 0) goto 9999

      if (start_name /= end_name) then
!write(6,*) '#'//start_name//'#'//end_name//'#'
          status = -200
          goto 9999
      endif

      return

 9999 continue
      call unstring(end_name)
      call unstring(start_name)

      end subroutine parse_element
!*******************************************************************************
      subroutine parse_element_starttag(buffer,c,handler,name,empty,status)

      implicit none
      character(*), intent(in) :: buffer
      integer, intent(inout) :: c
      integer, intent(out) :: status
      integer :: b,e,h,i
      logical :: has_attributes
      logical, intent(out) :: empty
      type(string), intent(out) :: name

      external handler


      b = c + 1
      c = b
      call parse_name(buffer,b,e,status)
      if (status /= 0) return
      i = e + 1


      call handler(HANDLER_START_ELEMENT,buffer(b:e),'',h)

      c = i
      has_attributes = .false.
      do while (.not. any(cbuffer(buffer,c,c) == (/ '/','>' /)))
          if (is_space(cbuffer(buffer,c,c))) then
              call parse_space(buffer,c)
          else
              if (is_space(cbuffer(buffer,c-1,c-1))) then
                  if (.not. has_attributes) then
                       has_attributes = .true.
                       call handler(HANDLER_START_ATTRIBUTES,'','',h)
                  endif
                  call parse_element_attribute(buffer,c,handler,status)
                  if (status /= 0) return
              else
                  status = -177
                  return
              endif
          endif
      enddo
      if (has_attributes) call handler(HANDLER_END_ATTRIBUTES,'','',h)

      select case(cbuffer(buffer,c,c))
      case('>')
          empty = .false.
          c = c + 1
      case('/')
          if (cbuffer(buffer,c+1,c+1) /= '>') then
              status = -22
              return
          endif
          c = c + 2
          empty = .true.
          call handler(HANDLER_EMPTY_ELEMENT,buffer(b:e),'',h)
      case default
          status = -23
          return
      end select

      name = cbuffer(buffer,b,e)

      end subroutine parse_element_starttag
!*******************************************************************************
      subroutine parse_element_attribute(buffer,c,handler,status)

      implicit none
      character(*), intent(in) :: buffer
      integer, intent(inout) :: c
      integer, intent(out) :: status
      integer :: b1,b2,e1,e2,h,i
      character :: d

      external handler


      i = bscan(buffer,c,achar(9)//achar(13)//achar(10)//achar(32)//'=')
      select case(i)
      case(0)
          status = -18
          return
      case default
          b1 = c
          e1 = i - 1
          c = i
          if (is_space(cbuffer(buffer,c,c))) call parse_space(buffer,c,status)
      end select

      if (is_space(cbuffer(buffer,c,c))) call parse_space(buffer,c,status)
      if (cbuffer(buffer,c,c) /= '=') then
          status = -19
          return
      endif

      c = c + 1
      if (is_space(cbuffer(buffer,c,c))) call parse_space(buffer,c,status)
      select case(cbuffer(buffer,c,c))
      case('"')
          d = '"'
      case("'")
          d = "'"
      case default
          status = -20
          return
      end select

      c = c + 1
      i = bscan(buffer,c,d)
      select case(i)
      case(0)
          status = -21
          return
      case default
          b2 = c
          e2 = i - 1
          c = i + 1
      end select

      call handler(HANDLER_ATTRIBUTE,cbuffer(buffer,b1,e1),  &
                   cbuffer(buffer,b2,e2),h)

      end subroutine parse_element_attribute
!*******************************************************************************
      subroutine parse_reference(buffer,c,handler,status)

      implicit none
      character(*), intent(in) :: buffer
      integer, intent(inout) :: c
      integer, intent(out) :: status
      integer :: b,e,h,i


      i = bindex(buffer,c,';')
      select case(i)
      case(0)
          status = -31
      case default
          b = c + 1
          e = i - 1
          c = i + 1
          status = 0
      end select

      call handler(HANDLER_REFERENCE,'',cbuffer(buffer,b,e),h)

      end subroutine parse_reference
!*******************************************************************************
      subroutine parse_chardata(buffer,c,handler,status)

      implicit none
      character(*), intent(in) :: buffer
      integer, intent(inout) :: c
      integer, intent(out) :: status
      integer :: b,e,h,i

      external handler


      i = bscan(buffer,c,'<&')
      select case(i)
      case(0)
          status = -18
      case default
          b = c
          e = i - 1
          c = i
          status = 0
      end select

      call handler(HANDLER_CHARDATA,'',cbuffer(buffer,b,e),h)

      end subroutine parse_chardata
!*******************************************************************************
      subroutine parse_element_endtag(buffer,c,handler,name,status)

      implicit none
      character(*), intent(in)   :: buffer
      integer, intent(inout)     :: c
      type(string), intent(out)  :: name
      integer, intent(out)       :: status
      integer                    :: b,e,h,i

      external handler


      b = c + 2
      call parse_name(buffer,b,e,status)
      if (status /= 0) return

      c = e + 1
      call parse_whitespace(buffer,c,c)
      c = c + 1

      if (cbuffer(buffer,c,c) /= '>') then
          status = -188
          return
      else
          c = c + 1
          status = 0
      endif

      name = cbuffer(buffer,b,e)

      call handler(HANDLER_END_ELEMENT,cbuffer(buffer,b,e),'',h)

      end subroutine parse_element_endtag
!*******************************************************************************
      subroutine parse_cdata(buffer,c,handler,status)

      implicit none
      character(*), intent(in) :: buffer
      integer, intent(inout) :: c
      integer, intent(out) :: status
      integer :: b,e,h,i

      external handler


      i = bindex(buffer,c,']]>')
      select case(i)
      case(0)
          status = -19
      case default
          b = c + 9
          e = i - 1
          c = i + 3
          status = 0
      end select

      call handler(HANDLER_CDATA,'',cbuffer(buffer,b,e),h)

      end subroutine parse_cdata
!*******************************************************************************
      subroutine parse_pi(buffer,c,handler,status)

      implicit none
      character(*), intent(in) :: buffer
      integer, intent(inout) :: c
      integer, intent(out) :: status
      integer :: b,e,h,i

      external handler


      i = bindex(buffer,c+2,'?>')
      select case(i)
      case(0)
          status = -24
      case default
          b = c + 2
          e = i
          c = i + 2
          status = 0
      end select

      call handler(HANDLER_PI,'',cbuffer(buffer,b,e),h)

      end subroutine parse_pi
!*******************************************************************************
      subroutine parse_comment(buffer,c,handler,status)

      implicit none
      character(*), intent(in) :: buffer
      integer, intent(inout) :: c
      integer, intent(out) :: status
      integer :: b,e,h,i

      external handler


      i = bindex(buffer,c,'-->')
      select case(i)
      case(0)
          status = -5
      case default
          select case(cbuffer(buffer,i-1,i-1))
          case('-')
              status = -6
          case default
              if (index(cbuffer(buffer,c+4,i-2),'--') /= 0) then
                  status = -7
              else
                  status = 0
                  b = c + 4
                  e = i - 1
                  c = i + 3
              endif
          end select
      end select

      call handler(HANDLER_COMMENT,'',cbuffer(buffer,b,e),h)

      end subroutine parse_comment
!*******************************************************************************
      subroutine parse_dtd(buffer,c,handler,status)

      implicit none
      character(*), intent(in) :: buffer
      integer, intent(inout) :: c
      integer, intent(out) :: status
      integer :: b,e,h,i

      external handler


      status = 0
      b = c + 9
      i = bscan(buffer,c,'[>')
      if (i == 0) then
          status = -11
          return
      endif

      if (cbuffer(buffer,i,i) == '[') then
          c = i + 1
          i = bindex(buffer,c,']')
          if (i == 0) then
              status = -111
              return
          endif
          c = i + 1
          i = bindex(buffer,c,'>')
          if (i == 0) then
              status = -111
              return
          endif
      endif

      e = i - 1
      c = i + 1

      call handler(HANDLER_DTD,'',cbuffer(buffer,b,e),h)

!     call dtd_parser(cbuffer(buffer,b,e),status)

      end subroutine parse_dtd

!*******************************************************************************
! Parse an XML declaration. It first checks whether it looks like an
! XML declaration, as it is not compulsary to have one.
! If not, it will just return. Otherwise it will parse the declaration.
!*******************************************************************************

      subroutine parse_xml_declaration(buffer,c,handler,status)

      implicit none
      character(*), intent(in) :: buffer
      integer, intent(inout) :: c
      integer, intent(out) :: status
      integer :: b,e,h,i

      external handler


      if (cbuffer(buffer,c,c+4) == '<?xml' .and.  &
          is_space(cbuffer(buffer,c+5,c+5))) then
          call handler(HANDLER_START_XML_DECLARATION,'','',h)
          i = bindex(buffer,c,'?>')
          select case(i)
          case(0)
              status = -4
          case default
              status = 0
              b = c
              e = i + 1
              call handler(HANDLER_END_XML_DECLARATION,'',cbuffer(buffer,b,e),h)
              c = i + 2
          end select
      else
!         Not an XML declaration; return and continue
          status = -1111111
      endif


      end subroutine parse_xml_declaration
!*******************************************************************************
      subroutine parse_space(buffer,c,status)

      implicit none
      character(*), intent(in) :: buffer
      integer, intent(inout) :: c
      integer, intent(out), optional :: status
      integer :: l


      if (is_space(cbuffer(buffer,c,c))) then
          l = len(buffer)
          do while (c <= l)
              if (.not. is_space(cbuffer(buffer,c,c))) exit
              c = c + 1
          enddo
          if (present(status)) status = 0
      else
          if (present(status)) status = -13
      endif

      end subroutine parse_space
!*******************************************************************************
      pure function is_space(c)

      implicit none
      character(*), intent(in)  :: c
      logical                   :: is_space
      character(*), parameter   :: space(4) = (/ achar( 9), achar(10),  &
                                                 achar(13), achar(32)   /)
      integer                   :: i,l


      is_space = .false.
      l = len(c)
      i = 1
      do while (i <= l)
          if (.not. any(c(i:i) == space)) return
          i = i + 1
      enddo
      is_space = .true.

      end function is_space
!*******************************************************************************

      function bindex(buffer,c,string) result(i)

      implicit none
      character(*), intent(in) :: buffer,string
      integer, intent(in) :: c
      integer :: i


      i = index(buffer(c:),string)
      if (i /= 0) i = c + i - 1

      end function bindex

!*******************************************************************************

      function bscan(buffer,c,string) result(i)

      implicit none
      character(*), intent(in) :: buffer,string
      integer, intent(in) :: c
      integer :: i

      i = scan(buffer(c:),string)
      if (i /= 0) i = c + i - 1

      end function bscan

!*******************************************************************************

      pure function cbuffer(buffer,b,e)

      implicit none
      character(*), intent(in) :: buffer
      integer, intent(in)      :: b,e
      character(e-b+1)         :: cbuffer


      cbuffer = buffer(b:min(e,len(buffer)))

      end function cbuffer

!*******************************************************************************

      subroutine parse_name(buffer,b,e,status)

      implicit none
      character(*), intent(in)  :: buffer
      integer, intent(in)       :: b
      integer, intent(out)      :: e,status
      integer                   :: i,l


      l = len(buffer)
      status = 0

      i = ichar(buffer(b:b))
      if (.not. (is_letter(i) .or. index(':_',buffer(b:b)) /= 0)) then
          status = -17
          return
      endif
      e = b + 1
      do while(e <= l)
          if (.not. is_namechar(ichar(buffer(e:e)))) exit
          e = e + 1
      enddo
      e = e - 1

      end subroutine parse_name

!*******************************************************************************

      subroutine parse_whitespace(buffer,b,e)

      implicit none
      character(*), intent(in)  :: buffer
      integer, intent(in)       :: b
      integer, intent(out)      :: e
      integer                   :: l


      l = len(buffer)
      e = b
      do while(e <= l)
          if (.not. (is_whitespace(buffer(e:e)))) exit
          e = e + 1
      enddo
      e = e - 1

      end subroutine parse_whitespace

!*******************************************************************************

      end module xml_parser
